/*!
 * Copyright (C) 2005-2006 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://jtagomat.sourceforge.net/
 */

/*!
 * $Log: arm7tdmi.c,v $
 * Revision 1.2  2006/03/20 14:10:17  haraldkipp
 * ReverseBitOrder() function removed. Instead the new function TapDataInverse()
 * will now be called to load debug scan chain 1. New function IceMultiWrite()
 * supports multiword write.
 *
 * Revision 1.1  2006/01/09 09:16:52  haraldkipp
 * First check-in.
 *
 */

#include "tap.h"
#include "ice.h"

/*! \name TAP Instructions */
/*@{*/
/*! Selects the scan path register. */
#define TAP_SCAN_N         0x02
/*! Returns core to system state after Run-Test/Idle. */
#define TAP_RESTART        0x04
/*! Places the previously selected scan chain in core test mode. */
#define TAP_INTEST         0x0C
/*! Selects the ID register. */
#define TAP_IDCODE         0x0E
/*@}*/

/*! \brief Debug control register.
 *
 * This read/write register is 3 bits wide.
 */
#define ICE_DBG_CTRL        0
#define ICE_DBGACK_BIT      0
#define ICE_DBGRQ_BIT       1
#define ICE_INTDIS_BIT      2   /*!< Set to 1 to disable core interrupts. */

/*! \brief Debug status register.
 *
 * This read/write register is 5 bits wide.
 */
#define ICE_DBG_STAT        1
#define ICE_IFEN_BIT        2   /*!< IFEN core state. */
#define ICE_NMREQ_BIT       3   /*!< NMREQ core state. */
#define ICE_TBIT_BIT        4   /*!< TBIT core state. */

/*! Debug comms control. */
#define ICE_DCC_CTRL        4
/*! Debug comms data. */
#define ICE_DCC_DATA        5
/*! Watchpoint 0 address value. */
#define ICE_WP0_ADDR        8
/*! Watchpoint 0 address mask. */
#define ICE_WP0_ADDR_MASK   9
/*! Watchpoint 0 data value. */
#define ICE_WP0_DATA        10
/*! Watchpoint 0 data mask. */
#define ICE_WP0_DATA_MASK   11
/*! Watchpoint 0 control value. */
#define ICE_WP0_CTRL        12
/*! Watchpoint 0 control mask. */
#define ICE_WP0_CTRL_MASK   13
/*! Watchpoint 1 address value. */
#define ICE_WP1_ADDR        16
/*! Watchpoint 1 address mask. */
#define ICE_WP1_ADDR_MASK   17
/*! Watchpoint 1 data value. */
#define ICE_WP1_DATA        18
/*! Watchpoint 1 data mask. */
#define ICE_WP1_DATA_MASK   19
/*! Watchpoint 1 control value. */
#define ICE_WP1_CTRL        20
/*! Watchpoint 1 control mask. */
#define ICE_WP1_CTRL_MASK   21

#define ICE_WRITE   0x20

#define DCC_HOST_TXBSY   1
#define DCC_HOST_RXRDY   2

#define ARM_NOP             0xE1A00000  /* MOV R0, R0 */

#define ARM_REG_SP      13
#define ARM_REG_LR      14
#define ARM_REG_PC      15
#define ARM_REG_CPSR    16

#ifndef _BV
#define _BV(x) (1 << (x))
#endif

unsigned long ice_dsr;
unsigned long arm_regs[17];

/*! 
 * \brief Select an ARM7 scan chain.
 * 
 * On return the selected chain will be in test mode.
 *
 * \param sc Number of the scan chain to select. Scan chain 0 allows 
 *           access to the entire peripherie of the core. Scan chain 1
 *           is a subset of scan chain 0 and allows access to the data 
 *           bus and the breakpoint signal. The EmbeddedICE-RT logic
 *           can be accessed via scan chain 2.
 */
static void Arm7SelectScanChain(unsigned char sc)
{
    static unsigned char current = 255;

    if (sc != current) {
        if (sc != 255) {
            /* Write the number of the scan chain into the scan path 
               select register. */
            TapInstruction(TAP_SCAN_N);
            TapData(0, &sc, 4);

            /* Put the selected scan chain in test mode. */
            TapInstruction(TAP_INTEST);
        }
        current = sc;
    }
}

/*!
 * \brief Access the core data bus and the BREAKPT signal.
 *
 * \param data  Contains the value to put on the data bus.
 * \param brkpt If set to 1, the next instruction will be executed at 
 *              system speed (MCLK).
 *
 * \return The value retrieved from the data bus.
 */
static unsigned long Arm7ScanChain1(unsigned long data, int brkpt)
{
    unsigned char wbuf[5];
    unsigned char rbuf[5];

    *((unsigned long *)wbuf) = data;
    wbuf[4] = (unsigned char) brkpt;

    /* Select the Debug Scan Chain */
    Arm7SelectScanChain(1);
    TapDataInverse(rbuf, wbuf, 33);
    TapNext(0);                 /* -> Run-Test/Idle for DCLK */

    return *((unsigned long *)rbuf);
}

/*!
 * \brief Access an EmbeddedICE-RT macro register.
 *
 * \param reg  Register number including the optional write flag (bit 5).
 * \param data Data to write into the macrocell register. Ignored if bit 5
 *             of the reg parameter is not set.
 *
 * \return If bit 5 of the reg parameter is not set, then the contents of
 *         the specified macrocell register is returned. Otherwise the
 *         returned value should be ignored.
 */
static unsigned long Arm7ScanChain2(int reg, unsigned long data)
{
    unsigned char buf[5];

    buf[0] = (unsigned char) data;
    buf[1] = (unsigned char) (data >> 8);
    buf[2] = (unsigned char) (data >> 16);
    buf[3] = (unsigned char) (data >> 24);
    buf[4] = (unsigned char) (reg);

    Arm7SelectScanChain(2);

    TapData(0, buf, 38);
    if (reg & ICE_WRITE) {
        return data;
    }
    TapData(buf, 0, 38);
    data = (unsigned long) (buf[0]) + ((unsigned long) (buf[1]) << 8) + 
        ((unsigned long) (buf[2]) << 16) + ((unsigned long) (buf[3]) << 24);
    return data;
}

/*!
 * \brief Execute a single instruction in test mode.
 *
 * Additional NOPs are pushed into the instruction pipeline to
 * move the instruction into the execute stage.
 *
 * \param opcode The instruction code to execute.
 */
static void Arm7TestExec(unsigned long opcode)
{
    /* Push instruction into the pipeline. */
    Arm7ScanChain1(opcode, 0);
    /* Push a NOP to move instruction into the decode stage. */
    Arm7ScanChain1(ARM_NOP, 0);
    /* Push a NOP to move instruction into the execute stage. */
    Arm7ScanChain1(ARM_NOP, 0);
}

/*!
 * \brief Set a given number of ARM CPU registers.
 *
 * \param regs The number of registers to set starting from R0.
 * \param tab  Pointer to an array containing the values to set.
 */
static void Arm7WriteRegs(int regs, unsigned long *tab)
{
    int i;
    unsigned long mask = 1;

    /* Set LDM mask. */
    for (i = 1; i < regs; i++) {
        mask |= (1 << i);
    }

    /* Execute LDMIA LR, {R0-Ri}. */
    Arm7TestExec(0xE89E0000 | mask);

    /* Put the values on the bus. */
    for (i = 0; i < regs; i++, tab++) {
        Arm7ScanChain1(*tab, 0);
    }

    /* This cycle stores the data bus value in the registers, */
    Arm7ScanChain1(ARM_NOP, 0);

    /* The program counter requires two additional cycles. */
    if (mask & 0x8000) {
        Arm7ScanChain1(ARM_NOP, 0);
        Arm7ScanChain1(ARM_NOP, 0);
    }
}

/*!
 * \brief Set CPU register contents.
 *
 * \param reg  Register index.
 * \param data Value to store into the specified register.
 */
static void Arm7WriteReg(unsigned char reg, unsigned long data)
{
    /* Execute LDR R[index], [R14] */
    Arm7TestExec(0xE59E0000 | ((unsigned long) reg << 12));

    /* Put the value on the bus. */
    Arm7ScanChain1(data, 0);

    /* This cycle stores the data bus value in the register, */
    Arm7ScanChain1(ARM_NOP, 0);

    /* The program counter requires two additional cycles. */
    if (reg == 15) {
        Arm7ScanChain1(ARM_NOP, 0);
        Arm7ScanChain1(ARM_NOP, 0);
    }
}

/*!
 * \brief Read CPU register contents.
 *
 * \param reg  Register index.
 *
 * \return Contents of the specified register.
 */
static unsigned long Arm7ReadReg(unsigned char reg)
{
    /* STR R[index], [R14]. Bits 12..15 specify the index. */
    Arm7TestExec(0xE58E0000 | ((unsigned long) reg << 12));

    /* Get data from the bus. */
    return Arm7ScanChain1(ARM_NOP, 0);
}

/*!
 * \brief Save ARM CPU registers.
 *
 * \todo Currently only R0-R15 and CPSR are saved.
 */
static void Arm7SaveContext(void)
{
    int i;

    /* Execute STMIA R0, {R0-R15}. */
    Arm7TestExec(0xE880FFFF);

    /* Get register contents from the bus. */
    for (i = 0; i < 16; i++) {
        arm_regs[i] = Arm7ScanChain1(ARM_NOP, 0);
    }

    /* Correct PC contents, 3 debug instruction + pipeline. */
    arm_regs[ARM_REG_PC] -= 5 * 4;

    /* Execute MRS R0, CPSR */
    Arm7TestExec(0xE10F0000);

    /* Read the result. */
    arm_regs[ARM_REG_CPSR] = Arm7ReadReg(0);
}

/*!
 * \brief Initialze the ICE routines.
 *
 * If in debug state, this routine will save the CPU context.
 */
void IceInit(void)
{
    /* Read the debug status register. */
    ice_dsr = Arm7ScanChain2(ICE_DBG_STAT, 0);
    if ((ice_dsr & (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) == (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) {
        /* Save context. */
        Arm7SaveContext();
    }
}

/*!
 * \brief De-initialze the ICE routines.
 *
 * If in debug state, this routine will re-store the CPU context.
 */
void IceExit(void)
{
    if ((ice_dsr & 9) == 9) {
        /* Restore context. */
        Arm7WriteReg(0, arm_regs[ARM_REG_CPSR]);
        Arm7TestExec(0xE12FF000);

        Arm7WriteRegs(14, arm_regs);
        Arm7WriteReg(ARM_REG_PC, arm_regs[ARM_REG_PC]);
        //Arm7WriteReg(ARM_REG_LR, arm_regs[ARM_REG_LR]);
    }
}

/*!
 * \brief Reset ICE routines to initial state.
 *
 * Should be called on hardware reset of the target.
 */
void IceReset(void)
{
    ice_dsr = 0;
    Arm7SelectScanChain(255);
}

/*!
 * \brief Get the device's ID code.
 *
 * \return ID code.
 */
unsigned long IceIdCode(void)
{
    unsigned long rc;

    TapInstruction(TAP_IDCODE);
    TapData((unsigned char *) &rc, 0, 32);

    return rc;
}

/*!
 * \brief Enter debug state.
 *
 * If the CPU is running, it will be stopped and the context will
 * be saved.
 *
 * \return 0 on success, -1 otherwise.
 *
 * \todo Thumb mode is not supported yet.
 */
int IceEnterDebug(void)
{
    unsigned short tmo = 0;

    /* Read the debug status register. TODO: Twice? */
    Arm7ScanChain2(ICE_DBG_STAT, 0);
    Arm7ScanChain2(ICE_DBG_STAT, 0);
    if ((ice_dsr & (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) != (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) {
        ice_dsr = Arm7ScanChain2(ICE_DBG_STAT, 0);

        /* Break on all addresses. */
        Arm7ScanChain2(ICE_WP0_ADDR_MASK | ICE_WRITE, 0xFFFFFFFF);
        /* Break on all data patterns. */
        Arm7ScanChain2(ICE_WP0_DATA_MASK | ICE_WRITE, 0xFFFFFFFF);
        /* Detect instruction fetch. */
        Arm7ScanChain2(ICE_WP0_CTRL_MASK | ICE_WRITE, 0xF7);
        /* Enable watchpoint 0. */
        Arm7ScanChain2(ICE_WP0_CTRL | ICE_WRITE, 0x00000100);
        /* Wait until DBGACK and nMREQ are both set to 1. */
        while (((ice_dsr = Arm7ScanChain2(ICE_DBG_STAT, 0)) & 
            (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) != (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) {
            if (++tmo >= 1000) {
                return -1;
            }
        }

        /* Disable watchpoints. */
        Arm7ScanChain2(ICE_WP0_CTRL | ICE_WRITE, 0);

        /* Go to Run-Test/Idle. */
        TapIdle();

        Arm7SaveContext();
    }
    return 0;
}

/*!
 * \brief Exit debug state.
 *
 * This routine will restore the context and start the CPU.
 *
 * \return 0 on success, -1 otherwise.
 */
int IceExitDebug(void)
{

    /* Store CPU status in R0 */
    Arm7WriteReg(0, arm_regs[ARM_REG_CPSR]);

    /* Execute MSR cpsr, R0 */
    Arm7TestExec(0xE12FF000);

    /* Restore context. */
    Arm7WriteRegs(14, arm_regs);
    Arm7WriteReg(ARM_REG_PC, arm_regs[ARM_REG_PC]);

    /* Execute next instruction at MCLK. */
    Arm7ScanChain1(ARM_NOP, 1);

    /* Execute  B [PC-5] at system speed (2 for ldm + 2 instructions) */
    Arm7ScanChain1(0xEAFFFFFB, 0);
    Arm7ScanChain1(ARM_NOP, 0);

    /* Restart at system speed as soon as Run-Test Idle is reached. */
    TapInstruction(TAP_RESTART);
    //TapNext(0);                 /* Go to Run-Test/Idle */
    TapIdle();
    Arm7ScanChain2(ICE_DBG_STAT, 0);
    IceReset();

    return 0;
}

/*!
 * \brief Get contents of a specified memory location.
 *
 * \param addr Memory location to read from.
 * \param data Pointer to the variable which receives the contents.
 *
 * \return 0 on success, -1 otherwise.
 *
 * \todo In order to speed up the transfer a routine for reading multiple
 *       locations is required.
 */
int IceMemoryRead(unsigned long addr, unsigned long *data)
{
    unsigned short tmo = 0;

    /* Load the memory address in r0. */
    Arm7WriteReg(0, addr);

    /* Make sure that the pipe is clean. */
    Arm7ScanChain1(ARM_NOP, 0);

    /* Execute next instruction at MCLK. */
    Arm7ScanChain1(ARM_NOP, 1);

    /* LDR R1, [R0], #4 */
    Arm7ScanChain1(0xE4901004, 0);

    Arm7ScanChain1(ARM_NOP, 0);

    /* Restart at system speed as soon as Run-Test Idle is reached. */
    TapInstruction(TAP_RESTART);
    TapNext(0);                 /* Go to Run-Test/Idle */

    /* Wait until DBGACK and nMREQ are both set to 1. */
    while ((Arm7ScanChain2(ICE_DBG_STAT, 0) & (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) != (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) {
        if (tmo++ >= 1000) {
            return -1;
        }
    }

    /* Read the result. */
    *data = Arm7ReadReg(1);

    return 0;
}

/*!
 * \brief Set contents of a specified memory location.
 *
 * \param addr Memory location to write to.
 * \param data Value to write.
 *
 * \return 0 on success, -1 otherwise.
 */
int IceMemoryWrite(unsigned long addr, unsigned long data)
{
    unsigned short tmo = 0;

    /* Load the memory address in r0. */
    Arm7WriteReg(0, addr);

    /* Load the memory contents in r1. */
    Arm7WriteReg(1, data);

    /* Make sure that the pipe is clean. */
    Arm7ScanChain1(ARM_NOP, 0);

    /* Execute next instruction at MCLK. */
    Arm7ScanChain1(ARM_NOP, 1);

    /* STR R1, [R0], #4 */
    Arm7ScanChain1(0xE4801004, 0);

    Arm7ScanChain1(ARM_NOP, 0);

    /* Restart at system speed as soon as Run-Test Idle is reached. */
    TapInstruction(TAP_RESTART);
    TapNext(0);                 /* Go to Run-Test/Idle */

    /* Wait until DBGACK and nMREQ are both set to 1. */
    while ((Arm7ScanChain2(ICE_DBG_STAT, 0) & (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) != (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) {
        if (tmo++ >= 1000) {
            return -1;
        }
    }
    return 0;
}

/*!
 * \brief Set contents of multiple memory locations.
 *
 * \param addr Memory location to write to.
 * \param data Values to write.
 *
 * \return 0 on success, -1 otherwise.
 */
int IceMultiWrite(unsigned long addr, unsigned long *data)
{
    unsigned short tmo = 0;

    Arm7WriteRegs(12, data);


    /* Load the memory address in r0. */
    Arm7WriteReg(14, addr);

    /* Make sure that the pipe is clean. */
    Arm7ScanChain1(ARM_NOP, 0);

    /* Execute next instruction at MCLK. */
    Arm7ScanChain1(ARM_NOP, 1);

    /* Execute STMIA LR, {R0-R11}. */
    Arm7ScanChain1(0xE88E0FFF, 0);

    Arm7ScanChain1(ARM_NOP, 0);

    /* Restart at system speed as soon as Run-Test Idle is reached. */
    TapInstruction(TAP_RESTART);
    TapNext(0);                 /* Go to Run-Test/Idle */

    /* Wait until DBGACK and nMREQ are both set to 1. */
    while ((Arm7ScanChain2(ICE_DBG_STAT, 0) & (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) != (_BV(ICE_DBGACK_BIT) | _BV(ICE_NMREQ_BIT))) {
        if (tmo++ >= 1000) {
            return -1;
        }
    }
    return 0;
}

/*!
 * \brief Send value to the debug communication channel.
 *
 * \param data Value to send to the target.
 *
 * \return 0 on success, -1 otherwise.
 */
int IceComWrite(unsigned long data)
{
    unsigned int tmo = 0;

    /* Wait until target read the previously sent data. */
    while (Arm7ScanChain2(ICE_DCC_CTRL, 0) & DCC_HOST_TXBSY) {
        if (tmo++ >= 5000) {
            return -1;
        }
    }
    Arm7ScanChain2(ICE_DCC_DATA | ICE_WRITE, data);

    return 0;
}

/*!
 * \brief Receive value from the debug communication channel.
 *
 * \param data Pointer to the variable which will received the value
 *             sent by the target.
 *
 * \return 0 on success, -1 otherwise.
 */
int IceComRead(unsigned long *data)
{
    unsigned int tmo = 0;

    /* Wait until target wrote data. */
    while ((Arm7ScanChain2(ICE_DCC_CTRL, 0) & DCC_HOST_RXRDY) == 0) {
        if (tmo++ >= 5000) {
            return -1;
        }
    }
    *data = Arm7ScanChain2(ICE_DCC_DATA, 0);

    return 0;
}


