/*!
 * Copyright (C) 2005-2006 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://jtagomat.sourceforge.net/
 */

/*!
 * $Log: comm.c,v $
 * Revision 1.3  2006/03/22 19:07:22  haraldkipp
 * Added a lot of comments to better explain the telegram state machine.
 * The input buffer is flushed before transmitting a new telegram.
 * Transmit retries are now handled in the SendTelegram()
 * routine. Retries are done only in case of NAK or timeout.
 * These changes make the telegram transfer more robust.
 *
 * Revision 1.2  2006/03/20 14:16:40  haraldkipp
 * Release 1.2.4.
 *
 * Revision 1.1  2006/01/09 09:16:52  haraldkipp
 * First check-in.
 *
 */

#include <avr/io.h>

#include "comm.h"

#ifdef __AVR_ATmega128__
#define UBRRL   UBRR0L
#define UDR     UDR0
#define UCSRA   UCSR0A
#define UCSRB   UCSR0B
#endif

#define STX     0x02
#define ETX     0x03
#define ACK     0x06
#define DLE     0x10
#define NAK     0x15

/*!
 * \brief Enable UART communication.
 */
void InitUart(void)
{
#ifdef __AVR_ATmega128__
    UBRRL = 7;
#else
    UBRRL = 3;
#endif
    UCSRB = (1 << RXEN) | (1 << TXEN);
}

static void SendChar(unsigned char ch)
{
    while((UCSRA & _BV(UDRE)) == 0);
    UDR = ch;
}

/*!
 * \brief Receive an STX/ETX/CKS telegram.
 *
 * \param buf Points to the buffer that receives the data bytes.
 * \param siz Size of the receiver buffer.
 *
 * \return The number of bytes received or -1 in case of an error.
 */
int RecvTelegram(unsigned char *buf, int siz)
{
    int rc = 0;
    unsigned char ch;
    unsigned char cs = 0; /* Makes the compiler happy. */
    unsigned short tmo = 0; /* Wait forever for first character. */
    unsigned char st = 0; /* State 0 waits for STX. */

    for (;;) {
        /* Wait until character received or timeout. */
        while((UCSRA & _BV(RXC)) == 0) {
            if (tmo && --tmo == 0) {
                return -1;
            }
        }
        ch = UDR;

        /* Update the checksum and enable timeout. */
        cs ^= ch;
        tmo = 0xFFFF;

        /*
         * State 1: Receiving data.
         */
        if (st == 1) {
            /* Unescaped STX, probably a new telegram. */
            if (ch == STX) {
                st = 0;
                /* State will be handled below. */
            }
            /* DLE is our escape character. */
            else if (ch == DLE) {
                /* Next character will not be interpreted. */
                st = 2;
                /* Revert checksum update. */
                cs ^= ch;
                /* Get a new character. */
                continue;
            }
            /* Unescaped ETX indicates end of telegram. */
            else if (ch == ETX) {
                /* Expect a checksum character. */
                st = 3;
                /* Get a new character. */
                continue;
            }
            /* Data characters will be handled below. */
        }

        /*
         * State 0: STX character expected.
         */
        if (st == 0) {
            if (ch == STX) {
                /* Got STX. Reset counter and checksum. */
                rc = 0;
                cs = 0;
                /* Now we are expecting data. */
                st++;
            }
            /* Respond to ctrl-V with version information. */
            else if (ch == 22) {
                char *vm = "Turtelizer " VERSION_STRING "\r\n";
                while (*vm) {
                    SendChar(*vm++);
                }
            }
            /* Get a new character. */
            continue;
        }

        /*
         * State 3: Checksum character expected.
         */
        if (st == 3) {
            if (cs) {
                /* Checksum error. */
                SendChar(NAK);
                rc = -1;
            }
            else {
                /* Valid telegram received. */
                SendChar(ACK);
            }
            /* Return result to the caller. */
            break;
        }

        /* Check for buffer overflow. */
        if (rc >= siz) {
            /* Return error to the caller. */
            rc = -1;
            break;
        }

        /* Store data character in buffer. */
        buf[rc] = ch;
        rc++;
        /* If last character was escaped, go back to normal. */
        st = 1;
    }
    return rc;
}

/*!
 * \brief Send an STX/ETX/CKS telegram.
 *
 * \param buf Points to the buffer that contains the data bytes.
 * \param siz Number of bytes in the transmit buffer.
 *
 * \return 0 on success or -1 in case of an error.
 */
int SendTelegram(const unsigned char *buf, int len)
{
    int rc;
    unsigned short tmo;
    unsigned char ch;
    unsigned char cs;
    unsigned char retry = 2; /* Three attempts. */

    while (retry--) {

        /* Flush input buffer. */
        while(UCSRA & _BV(RXC)) {
            ch = UDR;
        }

        /* Send STX character first. */
        SendChar(STX);

        /* Send data bytes. */
        cs = 0;
        for (rc = 0; rc < len; rc++) {
            ch = buf[rc];
            /* Update the checksum. */
            cs ^= ch;
            /* Escape control characters in the data stream. */
            if (ch == STX || ch == ETX || ch == DLE) {
                SendChar(DLE);
            }
            SendChar(ch);
        }
        /* Send ETX and checksum characters. */
        SendChar(ETX);
        cs ^= ETX;
        SendChar(cs);

        /*
         * Wait for ACK/NAK response. 
         */
        rc = -1;
        tmo = 0xFFFF;
        while((UCSRA & _BV(RXC)) == 0) {
            if (tmo && --tmo == 0) {
                break;
            }
        }

        if (tmo) {
            /* Got character before timeout elapsed. */
            ch = UDR;
            /* Only ACK indicates a valid transfer. */
            if (ch == ACK) {
                rc = 0;
                break;
            }
            /* Anything but NAK isn't worth a retry. */
            if (ch != NAK) {
                break;
            }
        }
    }
    return rc;
}

