/*!
 * Copyright (C) 2005-2006 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://jtagomat.sourceforge.net/
 */

/*!
 * $Log: tap.c,v $
 * Revision 1.2  2006/03/20 14:12:16  haraldkipp
 * Some parameters changed from 16-bit to 8-bit values. This improves speed
 * and reduces code size on 8-bit targets, like the Turtelizer's ATmega8.
 * New function TapDataInverse() added, which improves scan chain 1 access.
 *
 * Revision 1.1  2006/01/09 09:16:52  haraldkipp
 * First check-in.
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include "tap.h"

#ifdef __AVR_ATmega128__

#define RST_BIT     3
#define TCK_BIT     4
#define TMS_BIT     5
#define TDO_BIT     6
#define TDI_BIT     7
#define JTAG_PORT   PORTF
#define JTAG_PIN    PINF
#define JTAG_DDR    DDRF

#elif defined(__AVR_ATmega8__)

#define RST_BIT     5
#define TCK_BIT     1
#define TMS_BIT     2
#define TDO_BIT     3
#define TDI_BIT     0
#define JTAG_PORT   PORTC
#define JTAG_PIN    PINC
#define JTAG_DDR    DDRC

#else
#error "Unknown target MCU"
#endif

static unsigned char tap_shadow = _BV(RST_BIT);

void TapDelay(unsigned short l)
{
    while (l--) {
        asm volatile("nop");
    }
}

/*!
 * \brief Set TDI bit.
 *
 * \param s Specifies wether TDI should be set (1) or cleared (0).
 */
void TapTdi(int s)
{
    if (s) {
        tap_shadow |= _BV(TDI_BIT);
    } else {
        tap_shadow &= ~_BV(TDI_BIT);
    }
}

/*!
 * \brief Get TDO bit.
 *
 * \return 1 if TDI is set, 0 otherwise.
 */
int TapTdo(void)
{
    if (bit_is_set(JTAG_PIN, TDO_BIT)) {
        return 1;
    }
    return 0;
}

/*!
 * \brief Go to next TAP state.
 *
 * \param s Specifies wether TMS should be set (1) or cleared (0).
 */
void TapNext(int s)
{
    if (s) {
        tap_shadow |= _BV(TMS_BIT);
    } else {
        tap_shadow &= ~_BV(TMS_BIT);
    }

    /* Rising edge at TCK. */
    JTAG_PORT = tap_shadow;
    JTAG_PORT = tap_shadow | _BV(TCK_BIT);
}

/*!
 * \brief Reset TAP controller.
 *
 * \param s Specifies wether TRST should be set (1) or cleared (0).
 */
void TapReset(int s)
{
    unsigned char val;

    val = JTAG_PORT;
    if (s) {
        val &= ~_BV(RST_BIT);
    } else {
        val |= _BV(RST_BIT);
    }
    JTAG_PORT = val;
}

/*!
 * \brief Read and write register.
 *
 * \param rbuf Points to the buffer that receives the read data.
 *             May be NULL if the caller is not interested.
 * \param wbuf Points to the buffer that contains the write data.
 *             If NULL, all outgoing bits are zero.
 * \param len  Number of bits to read and/or write.
 */
static void TapRegister(unsigned char *rbuf, const unsigned char *wbuf, unsigned char len)
{
    while (len) {
        unsigned char mask = 1;
        unsigned char val = 0;

        if (wbuf) {
            val = *wbuf++;
        }
        while (len && mask) {
            TapTdi(val & mask);
            TapNext(--len == 0);
            if (TapTdo()) {
                val |= mask;
            } else {
                val &= ~mask;
            }
            mask <<= 1;
        }
        if (rbuf) {
            *rbuf++ = val;
        }
    }
}

/*!
 * \brief Read and write register in reverse bit order.
 *
 * \param rbuf Points to the buffer that receives the read data.
 *             May be NULL if the caller is not interested.
 * \param wbuf Points to the buffer that contains the write data.
 *             If NULL, all outgoing bits are zero.
 * \param len  Number of bits to read and/or write.
 */
static void TapRegisterInverse(unsigned char *rbuf, const unsigned char *wbuf, unsigned char len)
{
    unsigned char mask = 1 << ((len - 1) & 0x07);
    unsigned char val;

    rbuf += len / 8;
    wbuf += len / 8;
    while (len) {
        val = 0;
        if (wbuf) {
            val = *wbuf--;
        }

        while (mask) {
            TapTdi(val & mask);
            TapNext(--len == 0);
            if (TapTdo()) {
                val |= mask;
            } else {
                val &= ~mask;
            }
            mask >>= 1;
        }
        mask = 0x80;

        if (rbuf) {
            *rbuf-- = val;
        }
    }
}

/*!
 * \brief Read and write data register.
 *
 * \param rbuf Points to the buffer that receives the read data.
 *             May be NULL if the caller is not interested.
 * \param wbuf Points to the buffer that contains the write data.
 *             If NULL, all outgoing bits are zero.
 * \param len  Number of bits to read and/or write.
 */
void TapData(unsigned char *rbuf, const unsigned char *wbuf, unsigned char len)
{
    /* Run-Test/Idle, Update-DR or Update-IR state. */
    TapNext(1);                 /* -> Select-DR-Scan */
    TapNext(0);                 /* -> Capture-DR */
    TapNext(0);                 /* -> Shift-DR */

    TapRegister(rbuf, wbuf, len);

    TapNext(1);                 /* Exit1-DR -> Update-DR */
}

/*!
 * \brief Read and write data register in reverse bit order.
 *
 * \param rbuf Points to the buffer that receives the read data.
 *             May be NULL if the caller is not interested.
 * \param wbuf Points to the buffer that contains the write data.
 *             If NULL, all outgoing bits are zero.
 * \param len  Number of bits to read and/or write.
 */
void TapDataInverse(unsigned char *rbuf, const unsigned char *wbuf, unsigned char len)
{
    /* Run-Test/Idle, Update-DR or Update-IR state. */
    TapNext(1);                 /* -> Select-DR-Scan */
    TapNext(0);                 /* -> Capture-DR */
    TapNext(0);                 /* -> Shift-DR */

    TapRegisterInverse(rbuf, wbuf, len);

    TapNext(1);                 /* Exit1-DR -> Update-DR */
}

/*!
 * \brief Store instruction code in instruction register.
 */
void TapInstruction(unsigned long code)
{
    /* Run-Test/Idle, Update-DR or Update-IR state. */
    TapNext(1);                 /* -> Select-DR-Scan */
    TapNext(1);                 /* -> Select-IR-Scan */
    TapNext(0);                 /* -> Capture-IR */
    TapNext(0);                 /* -> Shift-IR */

    TapRegister(0, (unsigned char *) &code, 4);

    TapNext(1);                 /* Exit1-IR -> Update-IR */
}

/*!
 * \brief Go to Run-Test/Idle state.
 */
void TapIdle(void)
{
    int i;

    /* Any -> Test-Logic-Reset */
    for (i = 0; i < 5; i++) {
        TapNext(1);
    }
    TapNext(0);                 /* -> Run-Test/Idle */
}

/*!
 * \brief Initialize TAP controller.
 *
 * Ethernut is using
 *  - PF4 for target TCK.
 *  - PF5 for target TMS.
 *  - PF6 for target TDO.
 *  - PF7 for target TDI.
 *
 * \return 0 on success, -1 otherwise.
 *
 * \todo Implementing the Linux version.
 */
void TapInit(void)
{
    unsigned char val;

#ifdef __AVR_ATmega128__
    /*
     * The JTAG disable bit must be written twice within four cycles.
     * Thus we have to disable interrupts.
     */
    val = MCUCSR | _BV(JTD);
    cli();
    MCUCSR = val;
    MCUCSR = val;
    sei();
#endif

    /*
     * TCK, TMS and TDI output signals to target.
     * TDO input signal from target.
     */
    JTAG_PORT = _BV(RST_BIT) | _BV(TCK_BIT) | _BV(TMS_BIT) | _BV(TDI_BIT) | _BV(TDO_BIT);
    val = JTAG_DDR;
    val |= _BV(RST_BIT) | _BV(TCK_BIT) | _BV(TMS_BIT) | _BV(TDI_BIT);
    val &= ~_BV(TDO_BIT);
    JTAG_DDR = val;

    TapIdle();
}

/*!
 * \brief De-initialize the TAP controller.
 */
void TapExit(void)
{
    unsigned char val;

    TapIdle();

#ifdef __AVR_ATmega128__
    val = MCUCSR & ~_BV(JTD);
    /* Write twice within four cycles. */
    cli();
    MCUCSR = val;
    MCUCSR = val;
    sei();
#endif

    /* Set all lines back to input (z-state). */
    val = JTAG_DDR;
    val &= ~(_BV(RST_BIT) | _BV(TCK_BIT) | _BV(TMS_BIT) | _BV(TDI_BIT) | _BV(TDO_BIT));
    JTAG_DDR = val;
}


