/*!
 * Copyright (C) 2005-2006 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://jtagomat.sourceforge.net/
 */

/*!
 * $Log: turtelizer.c,v $
 * Revision 1.3  2006/03/22 19:06:28  haraldkipp
 * Removed transmission retries as these are now directly handled
 * by SendTelegram().
 *
 * Revision 1.2  2006/03/20 14:11:07  haraldkipp
 * Enhanced protocol for multiword write added.
 *
 * Revision 1.1  2006/01/09 09:16:52  haraldkipp
 * First check-in.
 *
 */

#include <avr/io.h>
#include <string.h>

#ifdef __AVR_ATmega128__

#include <cfg/os.h>
#include <cfg/arch.h>
#include <dev/debug.h>

#include <stdio.h>
#include <io.h>

#endif

#include "comm.h"
#include "tap.h"
#include "ice.h"
#include "led.h"

#define JTAG_OPEN       'O'
#define JTAG_CLOSE      'C'
#define JTAG_RESET      'Z'
#define JTAG_ENTER      'E'
#define JTAG_EXIT       'X'
#define JTAG_ID         'I'
#define JTAG_REGREAD    'r'
#define JTAG_REGWRITE   'w'
#define JTAG_MEMREAD    'R'
#define JTAG_MEMWRITE   'W'
#define JTAG_MULTIWRITE 'M'
#define JTAG_COMREAD    'i'
#define JTAG_COMWRITE   'o'

#define FIRMWARE_START  '0'

#define IERR_JTAG           1
#define IERR_BAD_COMMAND    16

static unsigned char tele[256];

/*!
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoOpen(unsigned char *tele, int len)
{
    TapInit();
    IceInit();
    tele[0] = 0;
    tele[1] = (unsigned char)VERSION;
    tele[2] = (unsigned char)(VERSION >> 8);
    tele[3] = (unsigned char)(VERSION >> 16);
    tele[4] = (unsigned char)(VERSION >> 24);

    return 5;
}

/*!
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoClose(unsigned char *tele, int len)
{
    IceExit();
    TapExit();
    tele[0] = 0;
    return 1;
}

/*!
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoReset(unsigned char *tele, int len)
{
    TapReset(1);
    TapDelay(10);
    TapReset(0);
    TapDelay(10);
    TapIdle();
    IceReset();
    tele[0] = 0;

    return 1;
}

/*!
 *
 * - Command: JTAG_ID(1)
 * - Result: result(1) data(4)
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoIdCode(unsigned char *tele, int len)
{
    unsigned long val;

    val = IceIdCode();
    tele[0] = 0;
    memcpy(&tele[1], &val, 4);
    return 5;
}

/*!
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoEnter(unsigned char *tele, int len)
{
    unsigned long val;

    /*
     * For some unknown reason this fails often.
     * We try twice.
     */
    tele[0] = 0;
    if (IceEnterDebug()) {
        IceExitDebug();
        if (IceEnterDebug()) {
            tele[0] = IERR_JTAG;
        }
    }
    if (tele[0] == 0) {
        IceMemoryRead(0, &val); /* Fixes something? */
    }
    return 1;
}

/*!
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoExit(unsigned char *tele, int len)
{
    if (IceExitDebug()) {
        tele[0] = IERR_JTAG;
    }
    else {
        tele[0] = 0;
    }
    return 1;
}

/*!
 * - Command: JTAG_REGREAD(1) index(1)
 * - Result: result(1) data(4)
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoRegRead(unsigned char *tele, int len)
{
    unsigned char idx = tele[1];

    tele[0] = 0;
    memcpy(&tele[1], &arm_regs[idx], 4);
    return 5;
}

/*!
 * - Command: JTAG_REGWRITE(1) index(1) value(4)
 * - Result: result(1)
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoRegWrite(unsigned char *tele, int len)
{
    unsigned char idx = tele[1];

    memcpy(&arm_regs[idx], &tele[2], 4);
    tele[0] = 0;
    return 1;
}

/*!
 * - Command: JTAG_MEMREAD(1) addr(4)
 * - Result: result(1) data(4)
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoMemRead(unsigned char *tele, int len)
{
    int rc;
    unsigned long addr;
    unsigned long data;

    memcpy(&addr, &tele[1], 4);
    if (IceMemoryRead(addr, &data)) {
        tele[0] = IERR_JTAG;
        rc = 1;
    }
    else {
        tele[0] = 0;
        memcpy(&tele[1], &data, 4);
        rc = 5;
    }
    return rc;
}

/*!
 * - Command: JTAG_MEMWRITE(1) addr(4) data(4)
 * - Result: result(1)
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoMemWrite(unsigned char *tele, int len)
{
    unsigned long addr;
    unsigned long data;

    memcpy(&addr, &tele[1], 4);
    memcpy(&data, &tele[5], 4);
    if (IceMemoryWrite(addr, data)) {
        tele[0] = IERR_JTAG;
    }
    else {
        tele[0] = 0;
    }
    return 1;
}

/*!
 * - Command: JTAG_MULTIWRITE(1) addr(4) data[0](4) ...  data[11](4)
 * - Result: result(1)
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoMultiWrite(unsigned char *tele, int len)
{
    unsigned long addr;
    unsigned long data[12];

    memcpy(&addr, &tele[1], 4);
    memcpy(&data, &tele[5], 48);
    if (IceMultiWrite(addr, data)) {
        tele[0] = IERR_JTAG;
    }
    else {
        tele[0] = 0;
    }
    return 1;
}

/*!
 * - Command: JTAG_COMREAD(1)
 * - Result: result(1) data(4)
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoComRead(unsigned char *tele, int len)
{
    int rc;
    unsigned long data;

    if (IceComRead(&data)) {
        tele[0] = IERR_JTAG;
        rc = 1;
    }
    else {
        tele[0] = 0;
        memcpy(&tele[1], &data, 4);
        rc = 5;
    }
    return rc;
}

/*!
 * - Command: JTAG_MEMWRITE(1) addr(4) data(4)
 * - Result: result(1)
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoComWrite(unsigned char *tele, int len)
{
    unsigned long data;

    memcpy(&data, &tele[1], 4);
    if (IceComWrite(data)) {
        tele[0] = IERR_JTAG;
    }
    else {
        tele[0] = 0;
    }
    return 1;
}


/*!
 * - Command: FIRMWARE_START(1)
 * - Result: none
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 */
void DoFirmwareStart(unsigned char *tele, int len)
{
#ifdef __AVR_ATmega8__
    asm volatile ("rjmp bootstart");
#endif
}

/*!
 *
 * \param tele Pointer to the command telegram.
 * \param len  Number of valid bytes in the command telegram.
 *
 * \return Size of the response telegram.
 */
int DoUnknown(unsigned char *tele, int len)
{
    tele[0] = IERR_BAD_COMMAND;
    return 1;
}

/*
 * Main application routine.
 */
int main(void)
{
    int rc;

    InitUart();

#ifdef __AVR_ATmega128__
    /*
     * Register the UART device, open it, assign stdout to it and set
     * the baudrate.
     */
    NutRegisterDevice(&devDebug1, 0, 0);
    freopen("uart1", "w", stdout);
    {
        unsigned long baud = 115200;
        _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
    }
    puts("\nTurtelizer " VERSION_STRING);
#endif

    LedBlink(10);
    for (;;) {
        Led(0);
        rc = RecvTelegram(tele, sizeof(tele));
        Led(1);
        if (rc > 0) {
            switch(tele[0]) {
            case FIRMWARE_START:
                DoFirmwareStart(tele, rc);
                break;
            case JTAG_OPEN:
                rc = DoOpen(tele, rc);
                break;
            case JTAG_CLOSE:
                rc = DoClose(tele, rc);
                break;
            case JTAG_RESET:
                rc = DoReset(tele, rc);
                break;
            case JTAG_ENTER:
                rc = DoEnter(tele, rc);
                break;
            case JTAG_EXIT:
                rc = DoExit(tele, rc);
                break;
            case JTAG_ID:
                rc = DoIdCode(tele, rc);
                break;
            case JTAG_REGREAD:
                rc = DoRegRead(tele, rc);
                break;
            case JTAG_REGWRITE:
                rc = DoRegWrite(tele, rc);
                break;
            case JTAG_MEMREAD:
                rc = DoMemRead(tele, rc);
                break;
            case JTAG_MEMWRITE:
                rc = DoMemWrite(tele, rc);
                break;
            case JTAG_MULTIWRITE:
                rc = DoMultiWrite(tele, rc);
                break;
            case JTAG_COMREAD:
                rc = DoComRead(tele, rc);
                break;
            case JTAG_COMWRITE:
                rc = DoComWrite(tele, rc);
                break;
            default:
                rc = DoUnknown(tele, rc);
                break;
            }
            SendTelegram(tele, rc);
        }
    }
}
